/*
    lkarmafs.0.1.4: FUSE/libkarma-based filsystem for the Rio Karma player
    Copyright (C) 2005-2006  Enrique Vidal <evidal@iti.upv.es>

    FUSE: Filesystem in Userspace
    Copyright (C) 2001-2005  Miklos Szeredi <miklos@szeredi.hu>

    libkarma: Library to access the Rio Karma player through its LAN interface
    Copyright (c) Frank Zschockelt <libkarma@freakysoft.de> 2004-2005

    This program can be distributed under the terms of the GNU GPL.
    See the file COPYING.

    Usage: 
    MOUNT:   lkarmafs [options] <mount directory>
             (see lkarmafs -H and lkarmafs -h for available options)
    UNMOUNT: fusermount -u <mount point>
*/

#define VERSION "0.2.0"

#ifdef linux
/* For pread()/pwrite() */
#define _XOPEN_SOURCE 500
#endif

#include <fuse.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <stdlib.h>
#include <locale.h>
#include <langinfo.h>
#include <time.h>
#include <limits.h>
#include <sys/types.h>
#include <signal.h>
#include <sys/xattr.h>
#include <sys/vfs.h>    /* or <sys/statfs.h> */
#include <mcheck.h>     /* for mtrace related functions */
// #include <attr/xattr.h>

#include "lkarma.h"
#include "utf8.h"
#include "pathedit.h"

#define DEBUG 0
#define FPRINTF if(DEBUG) fprintf

// #define MODE_PROPERTY "play_count_limit"      /* deprecated chmod property */
#define MODE_PROPERTY "bpm"                                 /* used for chmod */
#define W_PERM 1
#define R_PERM 0

#define TR_FRM "/"                     /* default 'from' path edditing string */
#define TR_TO  "|"                      /* default  'to' path edditing string */
#define TR_SLASH '|'                                   /* replacement for '/' */
//      TR_ESC  '\\'  /* edditing string escape defined as '\\' in pathedit.h */
#define UNK_NAME "-UnKnown-"

#define OPTIONS_FMT  "G=%d T=%d U=%d W=%d Y=%d End=0\n"
#define OPTIONS_SIZE 26

#define CHECK(x) (x<0)?"failed":"ok"

                            /* Some common strings needed trhoughout the code */
static char   *TYPE="type",      *GENRE="genre",    *ARTIST="artist",
            *SOURCE="source",    *TITLE="title",     *CODEC="codec", 
            *UPLOAD="upload",     *TAXI="taxi",   *PLAYLIST="playlist", 
              *TUNE="tune",     *VORBIS="vorbis",      *OGG="ogg",
           *BITRATE="bitrate",    *CTIME="ctime", *DURATION="duration", 
            *LENGTH="length",    *OFFSET="offset", *RUPLOAD="/upload",
          *ROPTIONS="/options",   *RTAXI="/taxi", 
    *FID_GENERATION="fid_generation",           *PLAY_COUNT="play_count",
  *PLAY_COUNT_LIMIT="play_count_limit",          *PLAY_LAST="play_last";

static int karma;
static int IOlock=0;

static int USB=1;        /* whether the USB or the Ethernet interface is used */
static int noGlobalWriteLock=0;                /* lock write only when needed */
static int G=1;                                     /* use Genre in pathnames */
static int updAccTime=1;          /* update play_last (access time) for tunes */
static int noXattr=0;               /* no xattr (extended attributes) support */
static int writeNoTaxi=0;            /* allow write under tune/ and playlist/ */
static int writeSmalldb=1;                    /* write smalldb back to the RK */
static int dirPlaylists=0;                     /* Directory view of playlists */
static char *UnkName=UNK_NAME;           /* Dummy name for missing properties */
static char *editPathStr1=TR_FRM;                   /* 'from' edditing string */
static char *editPathStr2=TR_TO;                      /* 'to' edditing string */

/* might be needed eventually
static char *mountPoint=NULL;                  // path to the mount directory //
static char myDir[PATH_MAX];                      // curent working directory //
*/

static void kfs_destroy(void *x);

/* **** CODE TO UPDATE TAXI FILES USING THE OLD MODE_PROPERTY TO NEW bpm **** */
/* ******************** TO BE ELIMINATED IN NEXT RELEASE ******************** */
static int UPDATE_MODE_PROP=0;
/* **************************** END UPDATING CODE *************************** */
/* ************************************************************************** */

/* -------------------------------------------------------------------------- */
inline int setProp(uint32_t fid, char *key, char *value) {
  return lk_properties_set_property(fid, key, value);
}
/* -------------------------------------------------------------------------- */
inline int setPropCS(uint32_t fid, char *key, char *value) {
  int ret; char *aux;
  aux = utf8_from_codeset(value);
  ret = lk_properties_set_property(fid, key, aux);
  free(aux); return ret; 
}
/* -------------------------------------------------------------------------- */
inline int setPropTRCS(uint32_t fid, char *key, char *value) {
  int ret; char *aux;
  if(value) unescape(value, editPathStr1, editPathStr2);
  aux = utf8_from_codeset(value);
  ret = lk_properties_set_property(fid, key, aux);
  free(aux); return ret; 
}

/* -------------------------------------------------------------------------- */
inline char* getProp(uint32_t fid, char *key) {
  return lk_properties_get_property(fid,key);
}
/* -------------------------------------------------------------------------- */
inline char* getPropCS(uint32_t fid, char *key) {
  return utf8_to_codeset(lk_properties_get_property(fid, key));
}

/* -------------------------------------------------------------------------- */
inline uint32_t* searchCS(int andOr, uint32_t *in, char *key, char *data) {
  uint32_t *ret; char *aux=NULL;
  if(data) {
    if(strcmp(data, UnkName)) aux = utf8_from_codeset(data);
  }
  else aux="";
  ret = lk_properties_andOrSearch(andOr+2, in, key, aux);
  return ret;
}

/* -------------------------------------------------------------------------- */
inline uint32_t *searchTRCS(int andOr, uint32_t *in, char *key, char *data) {
  uint32_t *ret; char *aux=NULL; 
/*fprintf(stderr, "+++> %s : %s\n", key, data); */
  if(data) {
    if(strcmp(data, UnkName)) {
      aux = strdup(data);
      unescape(aux, editPathStr1, editPathStr2);
      aux = utf8_from_codeset(aux);
    }
  }
  else aux=strdup("");
/*fprintf(stderr, "+++> %s : %s : %s\n", key, data, aux); */
  ret = lk_properties_andOrSearch(andOr+2, in, key, aux);        /* andOr+2:  */
  if(aux) free(aux); return ret;                              /* exact search */
/*int j=0; if(ret) for(j=0; (ret[j])!=0; j++) continue; */
/*fprintf(stderr, "+++> %s : %s : %s :: %d : %d\n", key,data,aux,j,ret); */
}

/* -------------------------------------------------------------------------- */
// int synchronize_necessary(int rio) {
//     int ret;
//     char *device_generation;
//     char *p;
// /*  fprintf(stderr, "##> synchronize_necessary\n"); */
// 
//     device_generation = 
//            strdup(lk_properties_get_property(0, "device_generation"));
//     lk_karma_get_device_settings(rio, &p);
//     lk_properties_import(p);
//     free(p);
//     ret=strcmp(device_generation, 
//                lk_properties_get_property(0, "device_generation"));
//     free(device_generation);
//     return ret;
// }

/* -------------------------------------------------------------------------- */
inline int waitForIOlock(void) {
  int i=0;
  while(IOlock) {usleep(100); i++; if(i>100) return -EINTR;}
  IOlock = 1;
  return 0;
}

/* -------------------------------------------------------------------------- */
inline void releaseIOlock(void) {
  IOlock = 0;
}

/* -------------------------------------------------------------------------- */
inline void condRequestKarmaIOWlock(void) {
    if(noGlobalWriteLock)
      fprintf(stderr, "lk_karma_request_io_lock (W): %s\n", 
                      CHECK(lk_karma_request_io_lock(karma,IO_LOCK_W)));
}

/* -------------------------------------------------------------------------- */
inline int condReleaseAndRequestKarmaIOlock(void) {
  int timeout;
  if(noGlobalWriteLock) {
    if((timeout = waitForIOlock())) return timeout;
      fprintf(stderr, "lk_karma_release_io_lock (W): %s\n", 
                       CHECK(lk_karma_release_io_lock(karma)));
    fprintf(stderr,   "lk_karma_request_io_lock (R): %s\n", 
                       CHECK(lk_karma_request_io_lock(karma,IO_LOCK_R)));
    releaseIOlock();
//  if(synchronize_necessary(karma)) { char *psw=NULL;
//    lk_karma_get_all_file_details(karma, &psw);
//    lk_properties_import(psw);
//    free(psw);
//  }
  }
  return 0;
}

/* -------------------------------------------------------------------------- */
uint32_t *copyListFids(uint32_t *vc) {
  int sizec=0; uint32_t *vp=NULL;
  if(vc) {
    while(vc[sizec] != 0) sizec++;                              /* size of vc */
    vp = malloc((sizec+1)*sizeof(uint32_t));
    memcpy(vp, vc, (sizec+1)*sizeof(uint32_t));                 /* copy vp=vc */
    free(vc);
  }
  return vp;
}

/* -------------------------------------------------------------------------- */
typedef struct pathElements {
  char *all;     char *type;    char *plylst;  char *genre;  
  char *artist;  char *source;  char *title;   char *codec;
} pathElementsType;
/* .......................................................................... */
pathElementsType parsePath(const char *path) {
  int lon;
  pathElementsType P;
  char *type=NULL,   *plylst=NULL, *genre=NULL, 
       *artist=NULL, *source=NULL, *title=NULL, *codec=NULL;

  P.all = NULL;
  if((path) && (strcmp(path, "/") != 0)) {
    lon = strlen(path); P.all = malloc(lon+4); strcpy(P.all, path);

    if(dirPlaylists && (!strncmp(path, "/playlist/", 10))) {
                                                   type = strchr(P.all, '/');
      if(type)   {*(type) = '\0'; type++;        plylst = strchr(type,  '/');}
      if(plylst) {*(plylst)='\0'; plylst++; if(G) genre = strchr(plylst,'/');
                                            else artist = strchr(plylst,'/');}
      if(G&&genre) {*(genre)  = '\0'; genre++;   artist = strchr(genre, '/');}
      if(artist)   {*(artist) = '\0'; artist++;  source = strchr(artist,'/');}
      if(source)   {*(source) = '\0'; source++;   title = strchr(source,'/');}
      if(title)    {*(title)  = '\0'; title++;    codec = strrchr(title,'.');
        if(codec)  {*(codec)  = '\0'; codec++;}
	  type = "tune";
      }
    } else {                                       type = strchr(P.all, '/');
      if(type)     {*(type) = '\0'; type++; if(G) genre = strchr(type,  '/');
                                            else artist = strchr(type,  '/');}
      if(G&&genre) {*(genre)  = '\0'; genre++;   artist = strchr(genre, '/');}
      if(artist)   {*(artist) = '\0'; artist++;  source = strchr(artist,'/');}
      if(source)   {*(source) = '\0'; source++;   title = strchr(source,'/');}
      if(title)    {*(title)  = '\0'; title++;    codec = strrchr(title,'.');
        if(codec)  {*(codec)  = '\0'; codec++;}
      }
    }

    if(codec && strcmp(codec,OGG)==0) {
      P.all = realloc(P.all, lon+4); strcpy(codec, VORBIS);
    }
    if(type && (genre || (!G))     && 
       ((strcmp(type, TAXI)==0)    || 
        (strcmp(type, UPLOAD)==0)  ||
        (strcmp(type, PLAYLIST)==0 && !dirPlaylists)))
      {title = G?genre:artist; genre = artist = NULL;}
  }
/*printf("++> %s %s %s %s %s %s\n", type,genre,artist,source,title,codec); */
  P.type   = type;   P.plylst = plylst; P.genre = genre; 
  P.artist = artist; P.source = source; P.title = title; P.codec = codec;
  return P;
}

/* -------------------------------------------------------------------------- */
uint32_t findFid(const char *path) {
  uint32_t fid=0, *vc=NULL, *vp=NULL;
  pathElementsType P;

  P = parsePath(path);
/*printf("++> %s : %s : %s : %s : %s : %s : %s\n", */
/*P.type, P.plylst, P.genre, P.artist, P.source, P.title, P.codec); */

  if(P.all) {
    if (dirPlaylists && P.plylst && !P.title) {
      vp = searchCS(0, NULL,   TITLE,  P.plylst);
      vc = searchCS(1,   vp,   TYPE,   P.type);                   /* frees vp */
      if(vc) {fid = vc[0]; free(vc);}	else fid = 1;     /* fid=1 => no file */
    }
    else if(P.title) {
      if(P.type && 
         ((strcmp(P.type, UPLOAD)==0) || (strcmp(P.type, PLAYLIST)==0))) {
	vp       = searchTRCS(0, NULL, TITLE,  P.title);
	vc       = searchCS(1,   vp,   TYPE,   P.type);           /* frees vp */
      }else if (P.type && (strcmp(P.type, TAXI)==0)) {
	vp       = searchCS(0,   NULL, TITLE,  P.title);
	vc       = searchCS(1,   vp,   TYPE,   P.type);           /* frees vp */
      }else{
	vp       = searchTRCS(0, NULL, TITLE,  P.title);
	vc       = searchTRCS(1, vp,   SOURCE, P.source);         /* frees vp */
	vp       = searchTRCS(1, vc,   ARTIST, P.artist);         /* frees vc */
        if(G) vc = searchTRCS(1, vp,   GENRE,  P.genre);          /* frees vp */
        else  vc = copyListFids( vp);                             /* frees vp */
	vp       = searchTRCS(1, vc,   TYPE,   P.type);           /* frees vc */
	vc       = searchTRCS(1, vp,   CODEC,  P.codec);          /* frees vp */
      }
      if(vc) {fid = vc[0]; free(vc);}	else fid = 1;     /* fid=1 => no file */
    }
    free(P.all);
  }
  return fid;
}

/* -------------------------------------------------------------------------- */
int addProps(uint32_t fid, pathElementsType P, unsigned int mode) {
  time_t generation_time;
  int ret=0;
/*printf("++> addProp: %d %s %s\n", fid, P.type, P.title); */

  generation_time = time(NULL);
  ret =   setProp(fid,      TYPE,            P.type)   +
          setPropTRCS(fid,  TITLE,           P.title)  +
          setProp(fid,      LENGTH,          "0")      +
          setProp(fid,      FID_GENERATION,  simple_itoa(generation_time)) +
          setProp(fid,      CTIME,           simple_itoa(generation_time)) +
          setProp(fid,      PLAY_LAST,       simple_itoa(generation_time));

  if(strcmp(P.type, TAXI)==0) {
    ret += setProp(fid,     CODEC,           TAXI) +
           setProp(fid,     MODE_PROPERTY,   simple_itoa(mode));
  }else if(strcmp(P.type, PLAYLIST)==0) {
    ret += setProp(fid,     MODE_PROPERTY,   simple_itoa(mode));
  }else{ /* tune */
    ret += setPropCS(fid,   GENRE,           P.genre)  +
           setPropTRCS(fid, ARTIST,          P.artist) +
           setPropTRCS(fid, SOURCE,          P.source) +
           setProp(fid,     CODEC,           P.codec)  +
           setProp(fid,     LENGTH,          "0")      +
           setProp(fid,     DURATION,        "0")      +
           setProp(fid,     OFFSET,          "0")      +
           setProp(fid,     BITRATE,         "0")      +
           setProp(fid,     PLAY_COUNT,      "0")      +
           setProp(fid,     PLAY_COUNT_LIMIT,"0");
  }
  if(ret) return -1;
  return 0;
}

/* -------------------------------------------------------------------------- */
int checkMode(uint32_t fid, int perm) {
  char *aux=NULL;  int mask=0;
/*fprintf(stderr, "++> checkMode: %d %d\n", fid, perm); */
  /* user perm. control is left to FUSE mount option -o default_permissions */
  /* uid_t  uid = fuse_get_context()->uid;  if(!(getuid()==uid)) return 1;  */
  if(strcmp(getProp(fid, TYPE), TAXI)) return 0;
  if(perm == W_PERM) mask = 0200;
  if(perm == R_PERM) mask = 0400;
  aux = getProp(fid, MODE_PROPERTY);
  if(!(aux && (mask & atoi(aux)))) return 1;
  return 0;
}

/* -------------------------------------------------------------------------- */
static uint32_t *getPlayListContents( uint32_t PLfid) {
    int i, j, count;
    uint32_t fid, fidGeneration, *vp = NULL;
    playlist *pl;
    
    pl = lk_playlist_fid_read(PLfid);  count = lk_playlist_count(pl);
/*  fprintf(stderr, "--> Contents of playlist \"%s\" (fid=%d, %d entries)\n", */
/*                                                 pl->name, pl->fid, count); */
    vp = malloc(count*sizeof(uint32_t)+1);  j=0;
    for (i=0; i<count; i++, j++)
      {lk_playlist_get_item(pl, i, &fid, &fidGeneration); vp[j] = fid; }
    vp[j] = 0;
    return vp;
}

/* -------------------------------------------------------------------------- */
/* -------------------------------------------------------------------------- */
static int kfs_utime(const char *path, struct utimbuf *buf) {
    int ret=0;
    uint32_t fid;
    FPRINTF(stderr, "==> UTIME: %s %d %d\n",
            path, (int)buf->actime, (int)buf->modtime);

    fid = findFid(path); 
    if(!buf) ret = setProp(fid, CTIME, simple_itoa(time(NULL)));
    else {
/*    ret += setProp(fid, "fid_generation", simple_itoa(buf->modtime)); */
      if(updAccTime || (strcmp(getProp(fid, TYPE), TAXI) == 0)) {
	ret += setProp(fid, CTIME,     simple_itoa(buf->modtime));
	ret += setProp(fid, PLAY_LAST, simple_itoa(buf->actime));
      }
    }
    if(ret) return -EPROTO;
    return 0;
}

/* -------------------------------------------------------------------------- */
static int kfs_chmod(const char *path, mode_t mode) {
    int res;
    char *aux;
    uint32_t fid;
    FPRINTF(stderr, "==> CHMOD: %s %o %s\n",path,(int)mode,simple_itoa(mode));

    if(!strcmp(path,"/")) {
      if(chmod(path, mode)) return -errno; return 0;
    }
    fid = findFid(path);
    if(fid>255) {
      if(strcmp(getProp(fid, TYPE), TAXI)) return -EROFS;
      condRequestKarmaIOWlock();
      if(setProp(fid, MODE_PROPERTY, simple_itoa(mode))) return -EIO;
      aux = lk_properties_export(fid);
      res = lk_karma_update_file_details(karma, fid, aux);
      free(aux);
      if(condReleaseAndRequestKarmaIOlock()) return -EIO;
      return 0;
    }
    else return -EACCES;
}

/* -------------------------------------------------------------------------- */
#ifdef FUSE22
static int kfs_statfs(const char *path, struct statfs *stbuf) {
#define USBDEVICE_SUPER_MAGIC 0x9fa2                        /* man (2) statfs */
/*  __fsid_t fsid; fsid.__val[0] = fsid.__val[1] = 0; */    /* man (2) statfs */
#else
static int kfs_statfs(const char *path, struct statvfs *stbuf) {
#endif
#define MAXNAMELEN 256
#define BLOCKSIZE 131072 /* 262144 131072 32768 */
    int res;
    uint32_t n_files=0, highest_file_id=0;
    uint64_t s_size=0,  f_space=0;
    double size, freeSize;
    (void) path;
    FPRINTF(stderr, "==> STATFS: %s\n", path);

    res = lk_karma_get_storage_details(
                       karma, 0, &n_files, &s_size, &f_space, &highest_file_id);
    if(res) res = EIO;
    size = s_size / BLOCKSIZE;  freeSize = f_space / BLOCKSIZE; 

    stbuf->f_bsize  = BLOCKSIZE;         /* long, optimal transfer block size */
    stbuf->f_frsize = BLOCKSIZE;                  /* fsblkcnt_t fragment size */
    stbuf->f_blocks = size;         /* long, total data blocks in file system */
    stbuf->f_bfree  = freeSize;                    /* long, free blocks in fs */
    stbuf->f_bavail = freeSize;          /* long, free blocks avail to non-su */
    stbuf->f_files  = n_files;       /* long, total file nodes in file system */
    stbuf->f_ffree  = highest_file_id-n_files; /* long, free file nodes in fs */
#ifdef FUSE22
    stbuf->f_type    = USBDEVICE_SUPER_MAGIC;     /* long, type of filesystem */
/*  stbuf->f_fsid    = fsid; */                     /* fsid_t, file system id */
    stbuf->f_namelen = MAXNAMELEN;       /* long, maximum length of filenames */
#else
    stbuf->f_flag    = 0xffff;             /* unsigned long, mount mask flags */
    stbuf->f_namemax = MAXNAMELEN;      /*  long, maximum length of filenames */
#endif
    return (-res);
}

/* -------------------------------------------------------------------------- */
static int kfs_listxattr(const char *path, char *list, size_t size) {
    char theList[254]=
         "fid\0offset\0trailer\0duration\0length\0samplerate\0ctime\0"
         "options\0bpm\0rms\0stddev\0norma\0lisation\0play_last\0"
         "play_count_limit\0play_count\0tracknr\0file_id\0marked\0"
         "playlist\0profile\0type\0codec\0artist\0source\0title\0"
         "bitrate\0genre\0year\0rid\0fid_generation\0profiler_version\0";
    size_t len = 254;                                 /* strlen(theList) + 1; */
    FPRINTF(stderr, "==> LISTXATTR: %s %d %d\n", path, size, len);
    (void) path;
 
    if(noXattr) return 0;
    if(size) memcpy(list, theList, size);
    return len;
}

/* -------------------------------------------------------------------------- */
static int kfs_getxattr(const char *path, 
			const char *key, char *value, size_t size) {
    size_t ret=0;
    uint32_t fid;
    char *aux=NULL;
    FPRINTF(stderr, "==> GETXATTR: %s %s %d\n", path, key, size);

    if(noXattr) return 0;
    if(strstr(key, "system.") != 0) return 0;

    fid = findFid(path);

    if(fid>255) {
      if(checkMode(fid, R_PERM)) return -EACCES;
      aux = getPropCS(fid, (char*)key); if(aux) ret = strlen(aux) + 5;
/*    fprintf(stderr, "--> %d %d %s\n", fid, ret, aux); */
      if(size && aux) {
	strncpy(value, "user.", 6); strncat(value, aux, size-6);
	if(ret>size) {ret = size; errno = ERANGE;}
      }
      else if(size) ret = -EINVAL;
    }
    else ret = -ENOENT;
    free(aux);

    return ret;
}

/* -------------------------------------------------------------------------- */
static int kfs_setxattr(const char *path, const char *key, 
			const char *value, size_t size, int flags) {
    int res=0;
    uint32_t fid;
    char *myKey=NULL, *myValue=NULL, *title=NULL, *aux=NULL;
    FPRINTF(stderr, "==> SETXATTR: %s %s %d\n", path, key, size);
    (void) flags; 

    if(noXattr) return 0;

    fid = findFid(path);
    if(fid>255) {
      if(!((strcmp(getProp(fid, TYPE), TAXI)==0) && getProp(fid, ARTIST))
	 && (checkMode(fid, W_PERM))) return -EACCES;
      myKey = strdup(key); 
      myValue = malloc(size+1); myValue[0]='\0';
      strncat(myValue, value, size);

      /* trick to allow simultanous change of type and title */
      if((strcmp(myKey, TYPE)==0)         && 
	 ((strncmp(myValue, TUNE, 4)==0)  ||
	  (strncmp(myValue, TAXI, 4)==0)) &&
	 (myValue[4]=='/')) {
	title = myValue+5; myValue[4] = '\0';
      }

      condRequestKarmaIOWlock();

      if(setPropCS(fid, myKey, myValue)) return -EINVAL;
      if(title) {if(setPropCS(fid, TITLE, title)) return -EINVAL;}

      aux = lk_properties_export(fid);
      res = lk_karma_update_file_details(karma, fid, aux);

      free(myKey); free(myValue); free(aux);

      if(condReleaseAndRequestKarmaIOlock()) return -ENOENT;
      return 0;
    }
    else return -ENOENT;
}

/* -------------------------------------------------------------------------- */
static int kfs_removexattr(const char *path, const char *name) {
    uint32_t fid;
    FPRINTF(stderr, "==> REMOVEXATTR: %s %s\n", path, name);

    return -ENOTSUP;       /* Disabled for now! needs a new libkarma function */

    fid = findFid(path);
/*  if(lk_properties_del_property(fid, name))  */
      return -EPROTO;
    return 0;
}

/* -------------------------------------------------------------------------- */
static int kfs_getattr(const char *path, struct stat *stbuf) {
    int ret=0;
    uint32_t fid;
    char *aux;
    FPRINTF(stderr, "==> GETATTR: %s\n", path);

    memset(stbuf, 0, sizeof(struct stat));

    stbuf->st_uid   = getuid();         stbuf->st_gid   = getgid();
    stbuf->st_nlink = 1;                stbuf->st_size  = 0;
    stbuf->st_ctime = stbuf->st_mtime = stbuf->st_atime = 0;
    stbuf->st_mode  = S_IFDIR|0755;

    if(strcmp(path, "/") == 0) return 0;
    if((!writeNoTaxi) && (strcmp(path, RTAXI)))
      stbuf->st_mode = S_IFDIR|0555;

    if(!strcmp(path, ROPTIONS)) 
      {stbuf->st_mode =  S_IFREG|0644; stbuf->st_size = OPTIONS_SIZE; return 0;}

    if((writeNoTaxi) && (!strncmp(path, RUPLOAD, 7))) {
      if(strlen(path) == 7)
	stbuf->st_mode = S_IFDIR|(writeNoTaxi?0755:0555);
      else {
	stbuf->st_mode = S_IFREG|(writeNoTaxi?0644:0444);
	if ((fid = findFid(path)) <= 1) ret = -ENOENT;
      }
      return ret;
    }

    fid = findFid(path);
/*  fprintf(stderr,"--> getattr, fid = %d\n", fid); */
    if(fid>255) {
//    if(checkMode(fid, R_PERM)) return -EACCES;
//    stbuf->st_nlink = 1; 
//    if(strcmp(getProp(fid, TYPE), TAXI)) stbuf->st_mode = 037777700444;

      if((!dirPlaylists) || strcmp(getProp(fid, TYPE), PLAYLIST)) {
        if(strcmp(getProp(fid, TYPE), TAXI))
          stbuf->st_mode = S_IFREG|(writeNoTaxi?0644:0444);
        else { /* for chmod */
          aux = getProp(fid, MODE_PROPERTY);
          if(aux) stbuf->st_mode = atoi(aux);
          else   {stbuf->st_mode = S_IFREG|0400; /* kfs_chmod(path, 0400); */}
        }
      }
      if((aux=getProp(fid, LENGTH)))    stbuf->st_size  =  (off_t)atol(aux);
      if((aux=getProp(fid, CTIME)))     stbuf->st_mtime = (time_t)atol(aux);
      if((aux=getProp(fid, PLAY_LAST))) stbuf->st_atime = (time_t)atol(aux);
      else                              stbuf->st_atime = stbuf->st_mtime;
      stbuf->st_ctime = stbuf->st_atime;
      stbuf->st_blocks = (stbuf->st_size >> 9);

      if((!dirPlaylists) && (!strcmp(getProp(fid, TYPE), PLAYLIST))) {
        playlist *pl = lk_playlist_fid_read(fid);
        stbuf->st_size = (off_t)pl->length;
	if(getProp(fid, LENGTH)) setProp(fid, LENGTH, simple_itoa(pl->length));
      }
    }
    else if(fid == 1) ret = -ENOENT;
/*  fprintf(stderr,"--> End getattr, ret = %d mode=%o\n",ret,stbuf->st_mode); */
    return ret;
}

/* -------------------------------------------------------------------------- */
static int kfs_readdir(const char *path, void *buf, fuse_fill_dir_t filler,
		                   off_t offset, struct fuse_file_info *fi) {
#define REPMAX 16384
    int j, lname=0, lrep=1, ret=0;
    uint32_t fid,      *vc=NULL,     *vp=NULL;
    char *key=NULL,    *name=NULL,   *aux=NULL, *rcodec=NULL, rep[REPMAX]="\n";
    char *codec=NULL;
    pathElementsType P;
    (void) offset; (void) fi;
    FPRINTF(stderr, "==> READDIR: %s\n", path);

    filler(buf, ".", NULL, 0); filler(buf, "..", NULL, 0);
    if (!path) return -ENOENT;

    P.all=NULL;  P = parsePath(path);  codec = P.codec;
    /* if type is taxi or playlist:  [genre,] artist and source are useless */

/*  printf("readir> %s : %s : %s : %s : %s : %s : %s\n",  */
/*  P.all, P.type, P.plylst, P.genre, P.artist, P.source, P.title); */

    /* FIX ME:  Integrate this code with the general one for tunes below */
    if(dirPlaylists && (!strncmp(path, "/playlist/", 10))) { 
      /* genre = playlist name */
      fid = findFid(path); vp = getPlayListContents(fid);
      if((!P.artist) && ((!P.genre) || (P.genre[0]=='\0'))) {
        if(G) {vc = searchTRCS(1,vp, GENRE,    NULL); key = GENRE;}
        else  {vc = vp;                               key = ARTIST;}
      }else{
        if(G) {vc = searchTRCS(1,vp, GENRE,  P.genre);  vp = copyListFids(vc);}
        if(!P.artist) {
             vc = searchTRCS(1, vp, ARTIST,     NULL); key = ARTIST; 
        }else{
             vc = searchTRCS(1, vp, ARTIST, P.artist);  vp = copyListFids(vc);
          if(!P.source) {
     	     vc = searchTRCS(1, vp, SOURCE,     NULL); key = SOURCE;
     	  }else{
     	     vc = searchTRCS(1, vp, SOURCE, P.source);  vp = copyListFids(vc);
     	     vc = searchTRCS(1, vp, TITLE,      NULL); key = TITLE;
     	  }
        }
      }
/*    fprintf(stderr, "--> /playlist/: fid = %d key = %s\n", fid, key); */
    }
    else if((!P.type) || (P.type[0]=='\0')) {
      filler(buf, "options", NULL, 0); 
      if(writeNoTaxi) filler(buf, UPLOAD, NULL, 0);
      vc = searchTRCS(0, NULL, TYPE, NULL);            key = TYPE; 
    }else{
      vc = searchTRCS(0, NULL, TYPE, P.type);  vp = copyListFids(vc);    
      if((!P.artist) && ((!P.genre) || (P.genre[0]=='\0'))) {
	if((strcmp(P.type, TAXI)==0) || (strcmp(P.type, PLAYLIST)==0)) {
                 vc = copyListFids(vp);                key = TITLE;
	}else{
          if(G) {vc=searchTRCS(1,vp, GENRE,     NULL); key = GENRE;}
          else  {vc = vp;                              key = ARTIST;}
        }
      }else{
        if(G) {vc = searchTRCS(1,vp, GENRE,  P.genre);  vp = copyListFids(vc);}
        if(!P.artist) {
             vc = searchTRCS(1, vp, ARTIST,     NULL); key = ARTIST; 
        }else{
             vc = searchTRCS(1, vp, ARTIST, P.artist);  vp = copyListFids(vc);
          if(!P.source) {
     	     vc = searchTRCS(1, vp, SOURCE,     NULL); key = SOURCE;
     	  }else{
     	     vc = searchTRCS(1, vp, SOURCE, P.source);  vp = copyListFids(vc);
     	     vc = searchTRCS(1, vp, TITLE,      NULL); key = TITLE;
     	  }
        }
      }
    }
    if(vc) {
      for(j=0; (fid=vc[j])!=0; j++) if(fid>256) {
/*      fprintf(stderr, "--> %d %d\n",j, fid); */
        aux = getPropCS(fid, key);  
        if (aux && strlen(aux)) {
          lname = strlen(aux)+1; name = malloc(2*lname+6);
          if (!name) continue;
          name[0] = '\n'; memcpy(name+1, aux, lname+1);
        } else {
          lname = strlen(UnkName)+1; name = malloc(2*lname+6);
          if (!name) continue;
          name[0] = '\n'; memcpy(name+1, UnkName, lname+1);
        }
        free(aux);
/*      fprintf(stderr, "-1-> %d %d %s %s\n", j, fid, key, name+1); */

        if(P.type && strcmp(P.type, TAXI) && 
                  (strcmp(P.type, PLAYLIST) || dirPlaylists)) {
          escape(name+1, editPathStr1, editPathStr2); lname = strlen(name);
        }
        if (P.source) {
          codec = getPropCS(fid, CODEC);
          if (codec) {
            rcodec = strcmp(codec, VORBIS)?codec:OGG;
            name[lname] = '.'; memcpy(name+lname+1, rcodec, strlen(rcodec));
            lname += strlen(rcodec)+1;
          }
        }

        name[lname+1] = '\0';
        name[lname]   = '\n';
        /* if not prevously seen (to avoid reps.) */
        if (strstr(rep, name) != NULL) {
          continue;
        }
        name[lname]   = '\0';
/*      fprintf(stderr, "-2-> %d %d %s\n", j, fid, name+1); */
        if (filler(buf, name+1, NULL, 0)) break;

        name[lname]   = '\n';
        if(lrep+lname+1 < REPMAX) {
          memcpy(rep+lrep, name+1, lname+1);
          lrep += lname;
        } else
          fprintf(stderr, "*Warning: too many repeated names\n");
        free(name);
      }
      free(vc); ret = 0; 
    }
    else ret = -ENOENT;

/*  fprintf(stderr, "-3-> %d\n", ret); */
    if(codec) free(codec); if(P.all) free(P.all);
    return ret;
}

/* -------------------------------------------------------------------------- */
static int kfs_mknod(const char *path, mode_t mode, dev_t rdev) {
                         /* Only taxi files can be written, for the moment... */
    int ret=0;
    uint32_t fid=0;
    pathElementsType P;
    FPRINTF(stderr, "==> MKNOD: %s\n", path);
    (void) rdev;

    P = parsePath(path);
    condRequestKarmaIOWlock();
    fid = lk_properties_new_property();                            /* new fid */
    if(fid < 256) {
/*    fprintf(stderr, "**lk error.\n"); */
      if(P.all) free(P.all); return -ECONNABORTED;
    }

    if(P.title) {
      if(!writeNoTaxi) {
	if(strcmp(P.type, TAXI)) ret = EROFS;
	else if(addProps(fid, P, mode)) ret = EPROTO; // mode=S_IFREG|0644; //??
      }
      else if(addProps(fid, P, mode)) ret = EPROTO;
    }
    else ret = ENOENT;

/*  fprintf(stderr, "--> End kfs_mknod %s %d %d\n", P.title, ret, fid); */
    if(P.all) free(P.all);
    if(ret) lk_properties_del_property(fid); 
    return -ret;
}

/* -------------------------------------------------------------------------- */
static int kfs_open(const char *path, struct fuse_file_info *fi) {
    uint32_t fid=0;
    FPRINTF(stderr, "==> OPEN: %s %o\n", path, fi->flags);

    if(!strcmp(path,ROPTIONS)) {fi->fh=0; return 0;}

    fid = findFid(path);
/*  printf("++> Fid = %d %d %d %d\n", fid, O_RDONLY, O_WRONLY, O_RDWR); */
    if(fid == 1) return -ENOENT;
    if((fi->flags & 3) == O_RDONLY) {                            /* READ open */
      if(fid < 256) return -ENOENT;
      if(checkMode(fid, R_PERM)) return -EACCES;
      fi->fh = (unsigned long)fid;
    }
    else {                                                      /* WRITE open */
//    if(checkMode(fid, W_PERM)) return -EACCES; /* fails create with mod 444 */
      if((fid < 256) || (strcmp(getProp(fid, TYPE), TAXI))) {
	if(!writeNoTaxi) return -EACCES;
      }
      fi->fh = (unsigned long)fid | 0x80000000;  /* leftmost bit = 1 => WRITE */
    }
    return 0;
}

/* -------------------------------------------------------------------------- */
static int kfs_read(const char *path, char *buf, size_t size, off_t offset,
		                                 struct fuse_file_info *fi) {
    int ret=0, res=0, timeout=0, lkerrno;
    uint64_t got=0;
    uint32_t fid=0;
    char *myBuf=NULL;          /* lk_karma_read_file_chunk mallocs its buffer */
    char *aux=NULL, aux1[256];
    FPRINTF(stderr, "==> READ: %s %d %ld %ld\n", 
                     path, (int)fi->fh, (long)size, (long)offset);

    if(!strcmp(path, ROPTIONS)) {
      sprintf(aux1, OPTIONS_FMT, 
                    G, writeNoTaxi, updAccTime, writeSmalldb, dirPlaylists); 
      got = strlen(aux1); memcpy(buf, aux1, got); return got;
    }
    (void) path;

    if((timeout = waitForIOlock())) return timeout;
    if((fid = (uint32_t)fi->fh&0x7FFFFFFF) > 255) {
      lk_errors_autoprint(0);
      res = lk_karma_read_file_chunk(karma, offset, size, fid, &myBuf, &got);
      lk_errors_autoprint(1);
      if(res) {
	lkerrno = lk_errors_number();
	aux = getProp(fid, LENGTH);
/*      printf("++> %4d %4d %4ld %4d\n", (int)fid,(int)got,atol(aux),lkerrno);*/
	if((lkerrno==E_FAILEDREQ) && (atol(aux)-offset))
	  {fprintf(stderr, "%s\n", lk_errors_numberstr(lkerrno)); ret = -EIO;}
      }
      else {
        /* patch to deal with inconsistent sizes of old riocp-loaded files */
	if(got < size) {aux=getProp(fid, LENGTH); got = atol(aux) - offset;}
	ret = got;
      }
      memcpy(buf, myBuf, got);  free(myBuf);
/*    printf("--> %4d %4d %4d %4d\n", (int)offset, size, (int)got, ret); */
    }
    releaseIOlock();

    return ret;
}

/* -------------------------------------------------------------------------- */
static int kfs_write(const char *path, const char *buf, size_t size,
                           off_t offset, struct fuse_file_info *fi) {
                         /* Only taxi files can be written, for the moment... */
    unsigned int totalWriteSize=0;
    int ret=0, res=0, timeout=0;
    uint32_t fid;
    FPRINTF(stderr, "==> WRITE: %s %d %d %d\n", \
                     path, (int)fi->fh&0x7FFFFFFF, size, (int)offset);

    #define optn(x) (strncmp(x, buf, 1)==0)
    if(!strcmp(path, ROPTIONS)) {
/*  fprintf(stderr, "--> WRITE: size=%d buf=%s\n", size, buf); */
      if      (optn("G")) G = G?0:1; 
      else if (optn("T")) writeNoTaxi = writeNoTaxi?0:1;
      else if (optn("U")) updAccTime = updAccTime?0:1;
      else if (optn("W")) writeSmalldb = writeSmalldb?0:1;
      else if (optn("Y")) dirPlaylists = dirPlaylists?0:1;
      else if (optn("E")) {kfs_destroy(""); kill(0,SIGINT);}
      return size;
    }

    (void) path;

    if((timeout = waitForIOlock())) return timeout;
    if((fid = ((uint32_t)fi->fh & 0x7FFFFFFF)) > 255) {
/*    fprintf(stderr, "--> fid = %d\n", fid); */
      res = lk_karma_write_file_chunk(karma, offset, size, fid, 0, buf);
/*    fprintf(stderr, "--> %s %4d %4d %d\n", path, fid, (int)offset, size); */
      if(res) {lk_properties_del_property(fid); ret = -EPROTO;}
      else {
        ret = size; totalWriteSize += (offset + size);
        res = setProp(fid, LENGTH, simple_itoa(totalWriteSize));
      }
    }
    releaseIOlock();

/*  fprintf(stderr, "--> End WRITE %4d %d\n", fid, ret); */
    return ret;
}

/* -------------------------------------------------------------------------- */
static int kfs_unlink(const char *path) {
    uint32_t fid=0;
    int ret=0;
    FPRINTF(stderr, "==> UNLINK: %s\n", path);
/* ....................... FIX ME (ETHERNET INTERFACE) ...................... */
/* `rm' crashes if used just after write, but not if write->umount->mount->rm */
/* .......................................................................... */

    fid = findFid(path);
    if(fid < 256) return -ENOENT;
/*  if(strcmp(getProp(fid,TYPE), TAXI)) return -EROFS; */
    if(checkMode(fid, W_PERM)) return -EACCES;

    condRequestKarmaIOWlock();

    ret = ret || lk_karma_delete_file(karma, fid);
    ret = ret || lk_properties_del_property(fid);
/*  fprintf(stderr, "--> fid = %d %d\n", fid, ret); */
    if(ret) return -EPROTO;

    return condReleaseAndRequestKarmaIOlock();
}

/* -------------------------------------------------------------------------- */
static int kfs_truncate(const char *path, off_t size) {
    uint32_t fid=0;
    int ret=0, theTime;
    char *buf="", *aux=NULL;
    FPRINTF(stderr, "==> TRUNCATE: %s: %d\n", path, (int)size);
/* ....................... FIX ME (ETHERNET INTERFACE) ...................... */
/* Only works after a fresh Rio-database download, with taxi files that were  */
/* in the filesystem before mounting.  Seems similar to the unlink bug.       */
/* .......................................................................... */
    if(size!=0) return -EFBIG;
    if(!strcmp(path, ROPTIONS)) return 0;

    fid = findFid(path);
    if(fid < 256) return -ENOENT;
    if(checkMode(fid, W_PERM)) return -EACCES;
    if(strcmp(getProp(fid, TYPE), TAXI)) return -EROFS;
/*  if((!writeNoTaxi) && (strcmp(getProp(fid,TYPE), TAXI))) return -EROFS;*/

    theTime = time(NULL);
    ret += setProp(fid, CTIME,     simple_itoa(theTime));
    ret += setProp(fid, PLAY_LAST, simple_itoa(theTime));
    if(ret) return -EPROTO;

    condRequestKarmaIOWlock();

    ret = ret || lk_properties_set_property(fid, LENGTH, "0");
    ret = ret || !(aux = lk_properties_export(fid));
    /* truncate failure happens here */ 
    ret = ret || lk_karma_delete_file(karma, fid);
    /* End truncate failure */ 
    ret = ret || lk_karma_write_file_chunk(karma, 0, 0, fid, 0, buf);
    ret = ret || lk_karma_update_file_details(karma, fid, aux);
/*  printf("--> %d %s ***%s***\n", ret, getProp(fid,"length"), aux); */
    free(aux);
    if(ret) return -EPROTO;

/*  printf("--> End truncate, ret = %d\n", ret); */
    return condReleaseAndRequestKarmaIOlock();
}

/* -------------------------------------------------------------------------- */
static int kfs_rename(const char *from, const char *to) {
    int ret=0;
    uint32_t fid=0;
    char *aux=NULL;
    pathElementsType T;
    FPRINTF(stderr, "==> RENAME: %s -> %s\n", from, to);

    fid = findFid(from);
    if(fid < 256) return -ENOENT;

    T = parsePath(to);
/*  printf("--> %s %s %s %s %s %s %s\n", */
/*          T.type, T.genre, T.artist, T.source, T.title, T.codec, T.all); */
/*  printf("--> %s -> %s\n", getProp(fid, "title"), titleTo); */

    if(!writeNoTaxi) {
      if(strcmp(getProp(fid, TYPE), TAXI) || (T.type && strcmp(T.type, TAXI)))
	return -EROFS;
    }

    condRequestKarmaIOWlock();
    ret   =    setPropTRCS(fid, TITLE,    T.title);
    if(writeNoTaxi) {
      if(strcmp(T.type, TAXI)==0) {
	ret  +=  setProp(fid,  CODEC, TAXI);
	if(strcmp(getProp(fid, TYPE), TAXI)) {
	  ret += setProp(fid,   MODE_PROPERTY, simple_itoa(S_IFREG|0444));
	}
      }
      else if(strcmp(T.type, PLAYLIST)==0){
	ret += setProp(fid,  PLAYLIST, "");
      }
      else if(strcmp(T.type, TUNE)==0){
	ret += setPropCS(fid,   GENRE,    T.genre)    +
	       setPropTRCS(fid, ARTIST,   T.artist)   +
               setPropTRCS(fid, SOURCE,   T.source)   +
               setPropTRCS(fid, TITLE,    T.title)    +
	       setProp(fid,     CODEC,    T.codec);    // +
//	       setProp(fid,     PLAY_COUNT_LIMIT, "0");
//	if(!(aux = getProp(fid, BITRATE)) || (strlen(aux)==0))
//             setProp(fid,     BITRATE, "fs128");  /* convenient default */
      }
      ret  +=  setProp(fid,     TYPE,     T.type);
    }
    aux = lk_properties_export(fid);  if(!aux) ret = -1;
    if(lk_karma_update_file_details(karma, fid, aux) || ret)
	ret = EPROTO;
    if(T.all) free(T.all); if(aux) free(aux);

    if(ret) return -ret;
    return condReleaseAndRequestKarmaIOlock();
}

/* -------------------------------------------------------------------------- */
static int kfs_flush(const char *path, struct fuse_file_info *fi) {
    uint32_t fid;
    int res=0, timeout=0, throughUpload;
    pathElementsType P;
    char *aux=NULL;
    FPRINTF(stderr, "==> FLUSH: %s %d %x\n",
	             path, (int)fi->fh & 0x7FFFFFFF, fi->flags);

    if(!strcmp(path, ROPTIONS)) return 0;
    (void) path;

    fid = (uint32_t)fi->fh & 0x7FFFFFFF;
    if(fid < 256) return -ENOENT;

    if(updAccTime || (strcmp(getProp(fid, TYPE), TAXI) == 0))
      res = setProp(fid, PLAY_LAST, simple_itoa(time(NULL)));

                                                           /* close READ open */
    if((fi->flags & 3) == O_RDONLY && ((uint32_t)fi->fh & 0x80000000) == 0) {
      if(res) return -EPROTO;
      return 0;
    }
                                                          /* close WRITE open */
    throughUpload = (!strcmp(getProp(fid, TYPE), UPLOAD));
    if((timeout = waitForIOlock())) return timeout;
    if(strcmp(getProp(fid, TYPE), TAXI)) {
      res = lk_rio_update_props_from_tags(karma, fid);
      if(res) {releaseIOlock(); return -EEXIST;}
    }
    res += setProp(fid, CTIME, simple_itoa(time(NULL)));
    if(!throughUpload) {
      P = parsePath(path);
      setPropCS  (fid, GENRE,  P.genre);
      setPropTRCS(fid, ARTIST, P.artist);
      setPropTRCS(fid, SOURCE, P.source);
      setPropTRCS(fid, TITLE,  P.title);
    }
    aux = lk_properties_export(fid);
    res += lk_karma_update_file_details(karma, fid, aux);
    releaseIOlock();
/*  fprintf(stderr, "--> %d %d ***%s***\n", fid, res, aux); */
    free(aux);
    if(res) return -EPROTO;

/*  fprintf(stderr, "--> End FLUSH %d\n", fid); */
    return condReleaseAndRequestKarmaIOlock();
}

/* -------------------------------------------------------------------------- */
static int kfs_release(const char *path, struct fuse_file_info *fi) {
    FPRINTF(stderr, "==> RELEASE: %s %d %x\n", path, (int)fi->fh, fi->flags);
    (void) fi; (void) path; return 0;
}

/* -------------------------------------------------------------------------- */
/*static void *kfs_init(void)
{ FPRINTF(stderr, "==> INIT\n"); }*/

/* -------------------------------------------------------------------------- */
static void kfs_destroy(void *x) {
    static int DONE=0;
    (void) x;
    FPRINTF(stderr, "==> DESTROY: x=%s, DONE=%d\n", (char *)x, DONE);

    if(DONE) return;
    DONE++;

    fprintf(stderr,   "lk_properties_save: %s\n", 
                 CHECK(lk_properties_save())); 
    if (writeSmalldb) 
      fprintf(stderr, "lk_karma_write_smalldb: %s\n",
                 CHECK(lk_karma_write_smalldb()));
    fprintf(stderr,   "lk_karma_release_io_lock: %s\n", 
                 CHECK(lk_karma_release_io_lock(karma)));
    fprintf(stderr,   "lk_karma_hangup: %s\n", 
                 CHECK(lk_karma_hangup(karma)));
    close(karma); lk_properties_destroy(); utf8_destroy();

    /* allow some time to finish pending (Ethernet) I/O calls */
    usleep(10000); if((!noGlobalWriteLock) && (!USB)) sleep(1);
    return;
}

/* -------------------------------------------------------------------------- */
static struct fuse_operations karma_oper = {
    .chmod       = kfs_chmod,
    .statfs      = kfs_statfs,
    .listxattr   = kfs_listxattr,
    .getxattr    = kfs_getxattr,
    .setxattr    = kfs_setxattr,
    .removexattr = kfs_removexattr,
    .getattr	 = kfs_getattr,
    .utime       = kfs_utime,
    .readdir	 = kfs_readdir,
    .mknod       = kfs_mknod,
    .open	 = kfs_open,
    .read	 = kfs_read,
    .write       = kfs_write,
    .truncate    = kfs_truncate,
    .unlink      = kfs_unlink,
    .rename      = kfs_rename,
    .flush       = kfs_flush,
    .release     = kfs_release,
    .destroy     = kfs_destroy,
/*  .init        = kfs_init, */
};

/* -------------------------------------------------------------------------- */
/* -------------------------------------------------------------------------- */
void usage(int complete) {
  printf("Usage:\n"
    "lkarmafs [fuse-mount_opts] [-h] \\\n"
    "         [-H|?] [-B] [-F] [-G] [-L] [-T] [-U] [-V] [-W] [-X] [-Y]\\\n" 
    "         [-A <path>|<hostname>|<IP>] [-C <userCodeset>] \\\n" 
    "         [-D <dummy-name>] [-E <tr-string1>] <tr-string2>] \\\n"
    "         [-P <passwd>]  <mountDirectory>\n");
  if(!complete) {printf("\n"); return;}
  printf("Options:\n"
    "  -h show FUSE and generic mount options\n"
    "  -H|? show this help (specific lkarmafs options and arguments)\n"
    "  -B get properties from ??1 files rather than \"smalldb\" (USB/OMFS)\n"
    "  -F force the files to be written even if they are duplicates\n"
    "  -G disable the 'genre' level in the filesystem hierarchy\n"
    "  -L don't use permanent write-lock\n"
    "  -T allow (limited) write permissions under playlist/ and tune/\n"
    "  -U don't update play_last (access time) for tune files\n"
    "  -V show current version\n"
    "  -W don't write \"smalldb\" upon unmounting (USB/OMFS interface)\n"
    "  -X don't use extended attributes (for tune properties)\n"
    "  -Y directory view of playlists\n");
  printf(
    "\n"
    "Using -L with the Ethernet interface, concurrent R/W may fail\n"
    "Change default -D name (\"-UnKnown-\") if it is used in some tune prop.\n"
    "\n"
    "Examples:\n"
    "  lkarmafs myMusicDir\n\n"
    "  lkarmafs -f -o allow_other,default_permissions \\\n"
    "           -P myKarma0 -C ISO8859-15 /mnt/karma\n\n"
    "  lkarmafs -f -o default_permissions -A /mnt/karma2 \\\n"
    "           -P myKarma0 -D \"MissingName\" -E \"/ \" \"+_\" ./karma\n"
    );
}

/* -------------------------------------------------------------------------- */
void currentKarma(int i, int *k) {
  static int karma;  if (i) karma=*k; else *k=karma;
}

/* -------------------------------------------------------------------------- */
double getUsedSize(int karma)
{
  uint32_t n_files, highest_file_id;
  uint64_t s_size, f_space;
  double size, freeSpace;

  n_files=s_size=f_space=highest_file_id=0;
  lk_karma_get_storage_details(karma, 0, 
                      &n_files, &s_size, &f_space, &highest_file_id);
  size = s_size; 
  freeSpace = f_space;
  return(size - freeSpace);
}

/* -------------------------------------------------------------------------- */
int initLibkarma(char *host, char *psw, char *user_codeset, 
		                        int useSmalldb, int writeDupes) {
#define ssdpRetry 3                     /* number of retries allowed for SSDP */
#define defaultCodeset "ISO8859-15"

  int dbChanged;
  int i, r;
  uint16_t port;
  char *usercodeset=NULL, *device=NULL, *settings;
/* ................................. FIX ME ................................. */
/* "Libkarma error: broken Rio header" after a while of mounted, if unused    */
/* .......................................................................... */
  if(user_codeset) usercodeset = strdup(user_codeset);

  lk_karma_write_dupes(writeDupes);
  if(useSmalldb) lk_karma_use_smalldb();

  lk_errors_autoprint(0);
  if(!host) { /* 'host' is a host name, an IP, or an USB mount directory */
    /* trying to find a USB/OMFS-mounted Karma */
    lk_errors_number();
    lk_mountSearch_discover(&device, &host);
    if(host) {
      fprintf(stderr, "Karma found: %s mounted on %s\n", device, host); USB=1;
    }else{
      fprintf(stderr, "No mounted Karma found; trying Ethernet.\n"); USB=0;
    }
  }
  if (!host) {
    /* trying to find an Ethernet-connected Karma */
    fprintf(stderr, "SSDP:");
    for(i=0; (i<=ssdpRetry)&&(host==NULL); i++) {
        lk_errors_number();
        lk_ssdp_discover(&host, &port);
        host?fprintf(stderr, " %s:%i\n", host, port)
            :fprintf(stderr, ".");
    }
    if(!host) fprintf(stderr, " no Karma found.\n");
  }
  if (!host) {
    lk_errors_p("*", " -- Can't connect to Rio Karma");
    usage(0);  if(psw) free(psw);  if(usercodeset) free(usercodeset);
    return -1;
  }
  karma = lk_karma_connect(host);  lk_errors_p("","");
  free(host);
  if(karma < 0) {
    usage(0); if(psw) free(psw);  if(usercodeset) free(usercodeset);
    return -1;
  }
  currentKarma(1,&karma);

  if(usercodeset){
//  lk_properties_init();
    utf8_set_codeset(usercodeset);
    if (lk_errors_p("", "; Trying locale...")) {
      free(usercodeset); usercodeset = NULL;
    }
  }
  if(usercodeset == NULL){
    if (! setlocale(LC_CTYPE, ""))
      fprintf(stderr,"* Warning: setlocale failed.\n");
    usercodeset = strdup(nl_langinfo(CODESET));
    if(strlen(usercodeset) == 0) {
      fprintf(stderr, "* Warning: nl_langinfo problem; "
	      "using default codeset %s\n", defaultCodeset);
      free(usercodeset); usercodeset = strdup(defaultCodeset);
    }
//  lk_properties_init();
    utf8_set_codeset(usercodeset);
  }
  fprintf(stderr, "Using %s codeset.\n", usercodeset);
  free(usercodeset); usercodeset = NULL;

  lk_errors_autoprint(1);                     /* Setting auto-print-errors ON */
  lk_properties_init();
  
  if(psw){
    fprintf(stderr, "lk_karma_authenticate: %i\n", 
	    lk_karma_authenticate(karma, psw));
    free(psw);
  }else fprintf(stderr, "lk_karma_authenticate: %i\n", 
		         lk_karma_authenticate(karma, ""));

  r = lk_karma_request_io_lock(karma,IO_LOCK_R); /* Read-IO_LOCK will be kept */
  fprintf(stderr, "lk_karma_request_io_lock (R): %s\n", CHECK(r)); /*always ON*/

  r+= lk_karma_get_device_settings(karma, &settings);
//fprintf(stderr, "lk_karma_get_device_settings: %s\n", CHECK(r));

  if(r < 0){
    printf("KarmaFS mount failed\n");
    lk_karma_hangup(karma);  close(karma);
    lk_properties_destroy(); utf8_destroy();
    return -1;
  }

  dbChanged = lk_properties_cache_obsolete(settings);
  if(dbChanged) {
    printf("Downloading Rio database... ");
    fflush(stdout);
    if(getUsedSize(karma)>3E9) printf("please wait... ");
    fflush(stdout);
  }
  free(settings);
  if(!r) lk_karma_load_database(karma);
  printf("%sKarmaFS mounted!\n", dbChanged?"done - ":"");

  if(!noGlobalWriteLock) {
//  if(synchronize_necessary(karma)) { char *psw=NULL;
//	lk_karma_get_all_file_details(karma, &psw);
//	lk_properties_import(psw);
//	free(psw);
//  }
    r = lk_karma_request_io_lock(karma,IO_LOCK_W);      /* also Write-IO_LOCK */
    fprintf(stderr,"lk_karma_request_io_lock (W): %s\n", CHECK(r));/*always ON*/
  }

  return 0;
}

/* -------------------------------------------------------------------------- */
int setEditPathStrings(int i, char *argv[]) { 

  unsigned int len;
  int ret;
  char *frm, *to;

  len = strlen(argv[i]);
  if (len != strlen(argv[i+1])) {
    fprintf(stderr, "ERROR: both -E edit strings must be of equal length\n");
    return -1;
  }
  if(editPathStr1 != TR_FRM) free(editPathStr1);
  if(editPathStr2 != TR_TO)  free(editPathStr2);
  editPathStr1 = malloc(len+2); memcpy(editPathStr1, argv[i++], len); 
  editPathStr2 = malloc(len+2); memcpy(editPathStr2, argv[i],   len);
  editPathStr1[len] = '\0';     editPathStr2[len] = '\0';

  ret = 0;
  for (frm = editPathStr1, to = editPathStr2; *frm != '\0'; frm++, to++) {
    if (*frm == *to) {
      fprintf(stderr, "ERROR: -E edit strings contain matching characters\n");
      return -1;
    } else if (*frm == '/') ret = 1;
    if (*to == '/') {
      fprintf(stderr, "ERROR: \'/\' is forbidden in 2nd edit string\n");
      return -1;
    } else if (*to == TR_ESC) {
      fprintf(stderr, "ERROR:\'%c\' is forbidden in 2nd edit string\n", TR_ESC);
      return -1;
    }
  }
  /* no '/' character found in editPathStr1 so we must add one */
  if (ret == 0) {
    editPathStr1[len] = '/';      editPathStr1[len+1] = '\0'; 
    editPathStr2[len] = TR_SLASH; editPathStr2[len+1] = '\0';
  }
  return 0;
}

/* -------------------------------------------------------------------------- */
int getConfFile(int myArgc, char **myArgv, int maxMyArg) {
  int i, len;
  fpos_t lastPos;
  FILE *fconf;
  char *arg, aux[PATH_MAX], confPath[PATH_MAX];
#define CONF_FILE_NAME "/.lkarmafsrc"
#define MAIN_CONF_PATH "/etc/lkarmafsrc"

  strncpy(confPath, getenv("HOME"), PATH_MAX); strcat(confPath, CONF_FILE_NAME);

  if(!(fconf = fopen(confPath, "r"))) {
    fprintf(stderr, "* No config file available:\n  "); perror(confPath);
    if(!(fconf = fopen(MAIN_CONF_PATH, "r"))) {
      fprintf(stderr,"* No config file available:\n  "); perror(MAIN_CONF_PATH);
      return myArgc;
    }
  }
  fgetpos(fconf, &lastPos);
  while (((i = fscanf(fconf, "%s ", aux)) > 0) && (myArgc<maxMyArg)) {
    len = strlen(aux);
/*  fprintf(stderr, "-+> %d %d %s\n", i, len, aux); */
    if((len==0) || (aux[0] == '#')) {
  	i = fscanf(fconf, "%*[^\n]\n");
  	continue;
    }
    arg = aux;
    if((aux[0]=='\'') && ((len==1) || (aux[len-1]!='\''))) {
  	if(fsetpos(fconf, &lastPos)) perror("fsetpos");
  	i = fscanf(fconf, "%*[']%[^']%*[']", aux);
    }
    else if((aux[0]=='"') && ((len==1) || (aux[len-1]!='"'))) {
  	if(fsetpos(fconf, &lastPos)) perror("fsetpos");
  	i = fscanf(fconf, "%*[\"]%[^\"]%*[\"]", aux);
    }
    else if((len>1) && (((aux[0]=='"')  && (aux[len-1]=='"')) ||
  			  ((aux[0]=='\'') && (aux[len-1]=='\'')))) 
  	{arg++; aux[len-1] = '\0';}

    myArgv[myArgc++] = strdup(arg);
    fgetpos(fconf, &lastPos);
/*  fprintf(stderr, "++> <<<<%s>>>>\n", myArgv[myArgc-1]); */
  }
  if((myArgc == maxMyArg) || (i != EOF)) {
    fprintf(stderr, "* Warning: bad config file %s ignored\n", confPath);
    myArgc = 1;
  }
  return myArgc;
}

/* ========================================================================== */
int main(int argc, char *argv[]) {
  int i, fuseArgc=0, myArgc=0, useSmalldb=1, writeDupes=0, ret;
  char **fuseArgv, **myArgv;
  char *usercodeset=NULL, *host=NULL, *psw=NULL;    /* Rio Karma paswrd */
/*char *usercodeset="ISO8859-15"; */
/*mtrace();  */
#define maxMyArg 64
#define option(x) ((strcmp(myArgv[i], (x)) == 0))
#define optArg(x) ((strcmp(myArgv[i], (x)) == 0) && (i++<myArgc-1))

  myArgv = malloc(maxMyArg*(sizeof(char*)));
  myArgv[0] = argv[myArgc++];

  myArgc = getConfFile(myArgc, myArgv, maxMyArg);

  for (i=1; (i<argc)&&(myArgc<maxMyArg); i++,myArgc++)
    myArgv[myArgc] = argv[i];
  if(myArgc == maxMyArg) fprintf(stderr, "* Warning: too many arguments\n");
/*for (i=0; i<myArgc; i++) fprintf(stderr, "--> %d **%s**\n", i, myArgv[i]); */

  fuseArgv = malloc(myArgc*(sizeof(char*)));

  for (i=0; i<myArgc; i++) {
      if(option("-h")) {
        fuseArgv[fuseArgc++] = myArgv[i];
	fuse_main(fuseArgc, fuseArgv, &karma_oper); return(0);
      }
      else if (option("-?")) {usage(0);}
      else if (option("-H")) {usage(1); return 0;}
      else if (option("-B")) useSmalldb=0;
      else if (option("-F")) writeDupes=1;
      else if (option("-G")) G = 0;
      else if (option("-L")) noGlobalWriteLock = 1;
      else if (option("-T")) writeNoTaxi = 1;
      else if (option("-U")) updAccTime = 0;
      else if (option("-V")) {printf("%s\n", VERSION); return 0;}
      else if (option("-W")) writeSmalldb=0;
      else if (option("-X")) noXattr = 1;
      else if (option("-Y")) dirPlaylists=1;
      /* CODE TO UPDATE TAXI FILES USING THE OLD MODE_PROPERTY TO NEW bpm */
      /* *************** TO BE ELIMINATED IN NEXT RELEASE *************** */
      else if (option("-Z")) UPDATE_MODE_PROP = 1;
      /* *********************** END UPDATING CODE ********************** */
      /* **************************************************************** */
      else if (optArg("-D")) {UnkName = strdup(myArgv[i]);}
      else if (optArg("-P")) {if(psw) free(psw); psw = strdup(myArgv[i]);}
      else if (optArg("-A")) {if(host) free(host); host = strdup(myArgv[i]);}
      else if (optArg("-C")) 
           {if(usercodeset) free(usercodeset); usercodeset= strdup(myArgv[i]);}
      else if (option("-E") && (i<myArgc-2))
           {if(setEditPathStrings(i+1, myArgv)) return -1; i+=2;}
      else fuseArgv[fuseArgc++] = myArgv[i];
  }

/*This code might be needed eventually
  mountPoint = strdup(fuseArgv[fuseArgc-1]);
  if(!getcwd(myDir, PATH_MAX)) perror("* Warning, could not get working dir");
*/
                                /* to allow mounting using /etc/fstab entries */
  if((fuseArgc > 2) && (fuseArgv[1][0] != '-') && (fuseArgv[2][0] != '-')) { 
    for (i=2; i<fuseArgc; i++) fuseArgv[i-1] = fuseArgv[i];
    fuseArgv[i]=NULL; fuseArgc--;
  }
/*fprintf(stderr, "++> fuseArgc = %d host = %s\n", fuseArgc, host); */
/*for (i=0; i<fuseArgc; i++) fprintf(stderr, "++> %d %s\n", i, fuseArgv[i]); */

  if(initLibkarma(host, psw, usercodeset, useSmalldb, writeDupes)) {
    fprintf(stderr, "ERROR: karma initialization failure\n");
    return -1;
  }

  /* CODE TO UPDATE TAXI FILES USING THE OLD MODE_PROPERTY TO NEW bpm */
  /* *************** TO BE ELIMINATED IN NEXT RELEASE *************** */
  if(UPDATE_MODE_PROP) {uint32_t fid=0, *vc=NULL; int j;
    #define MODE_PROP_OLD "play_count_limit"  /* deprecated chmod prop */
    fprintf(stderr, "***** Updating taxi modes from using the\n"
                    "      'play_count_limit' property to using 'bpm' *****\n");
    vc = searchCS(0, NULL, TYPE,  TAXI);
    for(j=0; (fid=vc[j])!=0; j++) {
      char *aux = getProp(fid, MODE_PROP_OLD);
      if(strcmp(aux, "0")) setProp(fid, MODE_PROPERTY, aux);
      else                 setProp(fid, MODE_PROPERTY, "33152");
      aux = lk_properties_export(fid);
      if (lk_karma_update_file_details(karma, fid, aux))
        fprintf(stderr, "Error updating taxi files mode property\n");
      free(aux);
    }
    free(vc);
  }
  /* *********************** END UPDATING CODE ********************** */
  /* **************************************************************** */

  ret = fuse_main(fuseArgc, fuseArgv, &karma_oper);
/*if(!ret) kfs_destroy(0); // ret==1 both if success and if mising mountpoint */
  kfs_destroy(""); free(myArgv); free(fuseArgv); return ret;

/*printf("--> END\n"); */
}
